package impl

import (
	"context"
	"errors"
	"fmt"
	"time"

	"gitlab.okta-solutions.com/mashroom/backend/common/api/grpcerr"
	"go.mongodb.org/mongo-driver/bson"

	"github.com/globalsign/mgo"
	invitation "gitlab.okta-solutions.com/mashroom/backend/invitation/api"
	notification "gitlab.okta-solutions.com/mashroom/backend/notification/api"
)

const (
	EmailSubject = "Verify your email for Mashroom"
	EmailBodyFmt = "To verify your email, use this code: %s"
	SmsBodyFmt   = "To verify your phone number, please enter this code: %s"

	EmailExpirationSeconds = 86400
	SmsExpirationSeconds   = 600 // change to 600 after autotesting
	SmsResendCooldown      = 10  // change to 10 after autotesting
)

var EmailVerificationUrlFmt = "https://" + SiteUrl + "/verify/email?code=%s"

func generateBody(t invitation.ContactValidationType, otp string) string {
	switch t {
	case invitation.ContactValidationType_EMAIL_VALIDATION:
		return fmt.Sprintf(EmailBodyFmt, otp)
	case invitation.ContactValidationType_PHONE_VALIDATION:
		return fmt.Sprintf(SmsBodyFmt, otp)
	}
	return ""
}

func expirationTimestamp(t invitation.ContactValidationType) time.Time {
	var dur int
	switch t {
	case invitation.ContactValidationType_EMAIL_VALIDATION:
		dur = EmailExpirationSeconds
	case invitation.ContactValidationType_PHONE_VALIDATION:
		dur = SmsExpirationSeconds
	default:
		dur = 0
	}
	return time.Now().Add(time.Second * time.Duration(dur))
}

func (server *serverImpl) CreateContactValidation(ctx context.Context, in *invitation.CreateContactValidationRequest) (*invitation.CreateContactValidationResponse, error) {
	if in.Type == invitation.ContactValidationType_UNKNOWN_VALIDATION {
		return nil, grpcerr.New(grpcerr.InvalidData, "unknown validation type")
	}

	var err error
	models, _ := server.repo.GetContactValidation(ctx, in.Type, in.AccountId, in.Key)

	now := time.Now()
	active := false
	var last time.Time
	for _, m := range models {
		if m.CreatedAt.After(last) {
			last = m.CreatedAt
		}
		if now.Sub(m.CreatedAt) > SmsExpirationSeconds*time.Second {
			_, _ = server.repo.contactValidations.DeleteOne(ctx, bson.M{"_id": m.Id})
		} else {
			active = true
			break
		}
	}

	if active && !in.GetForce() {
		return &invitation.CreateContactValidationResponse{
			Success: true,
		}, nil
	}

	if now.Sub(last) < SmsResendCooldown*time.Second {
		return &invitation.CreateContactValidationResponse{
			Success: false,
		}, nil
	}

	otp := generateOTP(in.Type)
	body := generateBody(in.Type, otp)

	model := &ContactValidationModel{
		AccountId: in.AccountId,
		Key:       in.Key,
		Otp:       otp,
		Body:      body,
		Type:      in.Type,
		ExpiresAt: now.Add(SmsExpirationSeconds * time.Second),
		CreatedAt: now,
	}

	fmt.Println(in.Type, in.AccountId, in.Key, "store", model)

	_, err = server.repo.CreateContactValidation(ctx, model)
	if err != nil {
		return nil, err
	}

	// dispatch notification
	switch model.Type {
	case invitation.ContactValidationType_EMAIL_VALIDATION:
		_, err := server.notification.SendEmail(ctx, &notification.SendEmailRequest{
			To:      in.Key,
			Subject: EmailSubject,
			Body:    body,
		})
		if err != nil {
			return nil, err
		}
	case invitation.ContactValidationType_PHONE_VALIDATION:
		_, err := server.notification.SendSms(ctx, &notification.SendSmsRequest{
			To:   in.Key,
			Body: body,
		})
		if err != nil {
			return nil, err
		}
	}

	return &invitation.CreateContactValidationResponse{
		Success: true,
	}, nil
}

func (server *serverImpl) VerifyContact(ctx context.Context, in *invitation.VerifyContactRequest) (*invitation.VerifyContactResponse, error) {
	models, err := server.repo.GetContactValidation(ctx, in.Type, in.AccountId, in.Key)
	fmt.Println(in.Type, in.AccountId, in.Key, in.Otp, "verif", models)
	if errors.Is(err, mgo.ErrNotFound) {
		return &invitation.VerifyContactResponse{Status: invitation.ContactValidationStatus_VALIDATION_NOT_FOUND}, nil
	} else if err != nil {
		return nil, err
	}

	now := time.Now()
	status := invitation.ContactValidationStatus_VALIDATION_FAILURE
	for _, m := range models {
		if m.ExpiresAt.Before(time.Now()) {
			_, _ = server.repo.contactValidations.DeleteOne(ctx, bson.M{"_id": m.Id})
			continue
		}
		if m.Otp == in.GetOtp() {
			if m.ExpiresAt.Before(now) {
				_, _ = server.repo.contactValidations.DeleteOne(ctx, bson.M{"_id": m.Id})
				if status != invitation.ContactValidationStatus_VALIDATION_CORRECT {
					status = invitation.ContactValidationStatus_VALIDATION_EXPIRED
				}
			} else {
				status = invitation.ContactValidationStatus_VALIDATION_CORRECT
			}
		}
	}
	if status != invitation.ContactValidationStatus_VALIDATION_CORRECT {
		if testOtp, ok := generateTestOTP(in.Key); ok {
			if testOtp == in.GetOtp() {
				status = invitation.ContactValidationStatus_VALIDATION_CORRECT
			}
		}
	}
	return &invitation.VerifyContactResponse{Status: status}, nil
}
