package impl

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"time"

	"github.com/golang/protobuf/ptypes"
	account "gitlab.okta-solutions.com/mashroom/backend/account/api"
	q "gitlab.okta-solutions.com/mashroom/backend/common/api/generic/query"
	"gitlab.okta-solutions.com/mashroom/backend/common/api/grpcerr"
	"gitlab.okta-solutions.com/mashroom/backend/common/api/invite"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/env"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/grpc"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/mongoquery"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/util"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/wrap"
	invitation "gitlab.okta-solutions.com/mashroom/backend/invitation/api"
	messaging "gitlab.okta-solutions.com/mashroom/backend/messaging/api"
	notification "gitlab.okta-solutions.com/mashroom/backend/notification/api"
	property "gitlab.okta-solutions.com/mashroom/backend/property/api"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

var (
	AccountServiceUrl = env.ResolveString("ACCOUNT_SERVICE_URL", "mashroom-account:10000")
	AccountService    = account.NewAccountServiceClient(grpc.Dial(AccountServiceUrl, "Account"))

	PropertyServiceUrl = env.ResolveString("PROPERTY_SERVICE_URL", "mashroom-property:10000")
	PropertyService    = property.NewPropertyServiceClient(grpc.Dial(PropertyServiceUrl, "Property"))

	ReferLandlordGoogleSheetURL = env.ResolveString("REFER_LANDLORD_GOOGLE_SHEET_URL",
		"https://script.google.com/macros/s/AKfycbwP_dSkJNXLMZTCXbIhEKu0mHzNr3O7D3aibnU3RtI56nx_iQ/exec")
)
var ErrNotAllowed = grpcerr.New(grpcerr.AuthError, "not allowed")
var ErrInvitationNotFound = grpcerr.New(grpcerr.NotFound, "invitation not found")

// ErrRequestEmpty is error occurred when required request data is empty.
var ErrRequestEmpty = grpcerr.New(grpcerr.InvalidData, "request empty")

// ErrBadHTTPResponseStatus is error occurred when some external provider returned no 200 status
var ErrBadHTTPResponseStatus = grpcerr.New(grpcerr.ProviderError, "http response status is not 200")

func (server *serverImpl) changeInvitationStatus(ctx context.Context, id string, invitationType invitation.InvitationType, user, email string, otp string, status invitation.InvitationStatus) (*invitation.InvitationActionResponse, error) {
	var model *InvitationModel
	var err error

	oid, err := primitive.ObjectIDFromHex(id)

	if len(id) > 0 && err == nil {
		model, err = server.repo.GetInvitationById(ctx, oid)
	} else {
		model, err = server.repo.GetInvitation(ctx, invitationType, email, otp)
	}
	if err != nil {
		return nil, err
	}

	switch model.Status {
	case invitation.InvitationStatus_PENDING:
		// sender may reject though
		if model.InviteeEmail != email {
			return nil, ErrNotAllowed
		}

		err = server.repo.ChangeInvitationStatus(ctx, model.Id, status)
		if err != nil {
			return &invitation.InvitationActionResponse{Error: invitation.InvitationActionError_UNKNOWN_ERROR, Success: false}, err
		}
		model.Status = status

		if server.bpms != nil {
			c, err := server.bpms.NewPublishMessageCommand().
				MessageName("invitation-reacted").
				CorrelationKey(model.Id.Hex()).
				VariablesFromMap(
					map[string]interface{}{
						"accepted": status == invitation.InvitationStatus_ACCEPTED,
						"user":     user,
					})
			if err == nil {
				_, err = c.Send(context.Background())
				if err != nil {
					log.WithError(err).Error()
				}
			} else {
				log.WithError(err).Error()
			}
		}

		if model.CallbackUrl != "" {
			client := grpc.Dial(model.CallbackUrl, "InviteCallback")
			inviteClient := invite.NewInviteServiceClient(client)
			switch status {
			case invitation.InvitationStatus_ACCEPTED:
				_, _ = inviteClient.InviteAccepted(context.TODO(), &invite.InviteAcceptedRequest{
					Invite:  model.Id.Hex(),
					Invitor: model.InviterID,
					Invitee: user,
				})
			case invitation.InvitationStatus_DECLINED:
				_, _ = inviteClient.InviteDenied(context.TODO(), &invite.InviteDeniedRequest{
					Invite:  model.Id.Hex(),
					Invitor: model.InviterID,
					Invitee: user,
				})
			}
		}

		return &invitation.InvitationActionResponse{
			Success: true,
			Info:    MakeInvitationInfo(model),
		}, nil
	case invitation.InvitationStatus_ACCEPTED:
		if status == invitation.InvitationStatus_DECLINED {
			return server.declineInvitation(ctx, id)
		}
		return &invitation.InvitationActionResponse{
			Error:   invitation.InvitationActionError_ALREADY_ACCEPTED,
			Success: false,
			Info:    MakeInvitationInfo(model),
		}, nil
	case invitation.InvitationStatus_DECLINED:
		return &invitation.InvitationActionResponse{
			Error:   invitation.InvitationActionError_ALREADY_DECLINED,
			Success: false,
			Info:    MakeInvitationInfo(model),
		}, nil
	default:
		return &invitation.InvitationActionResponse{
			Error:   invitation.InvitationActionError_UNKNOWN_ERROR,
			Success: false,
		}, nil
	}
}

func (server *serverImpl) declineInvitation(ctx context.Context, id string) (*invitation.InvitationActionResponse, error) {
	var model *InvitationModel
	var err error

	oid, err := primitive.ObjectIDFromHex(id)

	if len(id) > 0 && err == nil {
		model, err = server.repo.GetInvitationById(ctx, oid)
	} else {
		return nil, ErrInvitationNotFound
	}
	if err != nil {
		return nil, err
	}

	status := invitation.InvitationStatus_DECLINED
	switch model.Status {
	case invitation.InvitationStatus_PENDING:
		err = server.repo.ChangeInvitationStatus(ctx, model.Id, status)
		if err != nil {
			return &invitation.InvitationActionResponse{Error: invitation.InvitationActionError_UNKNOWN_ERROR, Success: false}, err
		}

		model.Status = status

		if server.bpms != nil {
			c, err := server.bpms.NewPublishMessageCommand().
				MessageName("invitation-reacted").
				CorrelationKey(model.Id.Hex()).
				VariablesFromMap(
					map[string]interface{}{
						"accepted": false,
						"user":     model.InviteeID,
					})
			if err == nil {
				_, err = c.Send(context.Background())
				if err != nil {
					log.WithError(err).Error()
				}
			} else {
				log.WithError(err).Error()
			}
		}

		if model.CallbackUrl != "" {
			client := grpc.Dial(model.CallbackUrl, "InviteCallback")
			inviteClient := invite.NewInviteServiceClient(client)
			_, _ = inviteClient.InviteDenied(context.TODO(), &invite.InviteDeniedRequest{
				Invite:  model.Id.Hex(),
				Invitor: model.InviterID,
				Invitee: model.InviteeID,
			})
		}

		if model.Type == invitation.InvitationType_LANDLORD_INVITATION || model.Type == invitation.InvitationType_MANAGER_TO_LANDLORD_INVITATION {
			_, _ = server.notification.SendEvent(ctx, &notification.SendEventRequest{
				User:     model.InviterID,
				Event:    notification.EventType_EVENT_LISTING_INVITATION_REJECTED_TENANT,
				Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_EMAILV2, notification.EventChannel_EVENT_CHANNEL_PLATFORM},
				Embeds: []*notification.EventEmbed{
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
						EmbedValue: model.InviteeID,
					},
				},
			})
		}

		if model.Type == invitation.InvitationType_MANAGER_INVITATION {
			_, _ = server.notification.SendEvent(ctx, &notification.SendEventRequest{
				User:     model.InviterID,
				Event:    notification.EventType_EVENT_LISTING_INVITATION_REJECTED_LANDLORD,
				Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_EMAILV2, notification.EventChannel_EVENT_CHANNEL_PLATFORM},
				Embeds: []*notification.EventEmbed{
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_INVITATION,
						EmbedValue: model.Id.Hex(),
					},
				},
			})
		}

		return &invitation.InvitationActionResponse{
			Success: true,
			Info:    MakeInvitationInfo(model),
		}, nil
	case invitation.InvitationStatus_ACCEPTED:
		switch model.Type {
		case invitation.InvitationType_MANAGER_INVITATION, invitation.InvitationType_LANDLORD_INVITATION, invitation.InvitationType_MANAGER_TO_LANDLORD_INVITATION:
			errResponse := &invitation.InvitationActionResponse{Error: invitation.InvitationActionError_UNKNOWN_ERROR, Success: false}
			// if property status Listed or lower status, manager user to change to landlord user
			err = server.repo.ChangeInvitationStatus(ctx, model.Id, status)
			if err != nil {
				log.Infof("declineInvitation ChangeInvitationStatus error %v\n", err)
				return errResponse, err
			}

			prop, err := PropertyService.Get(ctx, &property.GetRequest{
				Id:   model.PropertyID,
				User: model.InviterID,
			})
			if err != nil {
				log.Infof("declineInvitation get property error %v\n", err)
				return errResponse, err
			}

			people := prop.GetProperty().GetPeople()
			propertyStatus := prop.GetProperty().GetStatus()

			if isPropertyNotOffered(propertyStatus) {
				if people.GetManager() != people.GetLandlord() && people.GetLandlord().GetValue() != "" {
					people.Manager = people.GetLandlord()
					p, err := PropertyService.UpdateAndVerify(ctx, &property.UpdateAndVerifyRequest{
						User:        model.InviterID,
						Property:    prop.GetProperty(),
						AdminVerify: &property.AdminVerify{},
					})
					if err != nil {
						log.Infof("declineInvitation update property error %v\n", err)
						return errResponse, err
					}

					log.Infof("declineInvitation update property success %+v\n", *p.GetProperty())
				}

				return &invitation.InvitationActionResponse{
					Success: true,
					Info:    MakeInvitationInfo(model),
				}, nil
			}
			return &invitation.InvitationActionResponse{
				Error:   invitation.InvitationActionError_ALREADY_ACCEPTED,
				Success: false,
				Info:    MakeInvitationInfo(model),
			}, nil

		default:
			return &invitation.InvitationActionResponse{
				Error:   invitation.InvitationActionError_ALREADY_ACCEPTED,
				Success: false,
				Info:    MakeInvitationInfo(model),
			}, nil
		}
	case invitation.InvitationStatus_DECLINED:
		return &invitation.InvitationActionResponse{
			Error:   invitation.InvitationActionError_ALREADY_DECLINED,
			Success: false,
			Info:    MakeInvitationInfo(model),
		}, nil
	default:
		return &invitation.InvitationActionResponse{
			Error:   invitation.InvitationActionError_UNKNOWN_ERROR,
			Success: false,
		}, nil
	}
}

func isPropertyNotOffered(propertyStatus property.PropertyStatus) bool {
	return propertyStatus == property.PropertyStatus_LISTED || propertyStatus <= 4
}

func (server *serverImpl) CreateInvitation(ctx context.Context, req *invitation.InvitationResolution) (*invitation.InvitationActionResponse, error) {
	log.Infof("CreateInvitation request: %+v\n", req)

	// get invitee extra information, checking invitee account existing
	inviteeFirstName, inviteeLastName, inviteeID, inviteeEmail := getInvitationInviteeInfo(ctx, req.GetInvitee())

	// generate id for new invitation
	id := primitive.NewObjectID()

	// generate notification message body from request
	body := getInvitationMessageBody(req, id)

	invitationType := req.GetType()
	// create invitation model
	model := &InvitationModel{
		Id:              id,
		Type:            invitationType,
		Status:          invitation.InvitationStatus_PENDING,
		InviterID:       req.GetInviter().GetId(),
		InviteeID:       inviteeID,
		InviteeEmail:    inviteeEmail,
		InviteFirstName: inviteeFirstName,
		InviteLastName:  inviteeLastName,
		Otp:             generateEmailOTP(),
		Body:            body,
		PropertyID:      req.GetProperty().GetId(),
		CreatedAt:       time.Now(),
		ExpiresAt:       time.Now().Add(time.Hour * 24 * time.Duration(req.GetExpiresDays())),
		Process:         req.GetProcess(),
		Offer:           req.GetOffer(),
		CallbackUrl:     req.GetCallbackUrl(),
		Message:         req.GetMessage(),
	}

	// set DECLINED status to previous invitations of the same type and property
	switch invitationType {
	case invitation.InvitationType_LANDLORD_INVITATION, invitation.InvitationType_MANAGER_INVITATION, invitation.InvitationType_MANAGER_TO_LANDLORD_INVITATION:
		err := declinedExistsInvitations(ctx, server, model)
		if err != nil {
			return nil, err
		}
	}

	// insert into db (upsert actually)
	log.Infof("CreateInvitation create model =  %+v", model)
	_, err := server.repo.CreateInvitation(ctx, model)
	if err != nil {
		return nil, err
	}

	switch invitationType {
	case invitation.InvitationType_LANDLORD_INVITATION: // auto set landlord to property
		prop, err := PropertyService.Get(ctx, &property.GetRequest{
			User: req.GetInviter().GetId(),
			Id:   req.GetProperty().GetId(),
		})
		if err != nil {
			return nil, err
		}

		if inviteeID != "" { // if user account exist, set user as landlord
			isNeedPropertyUpdate := false
			people := prop.GetProperty().GetPeople()
			if people.GetLandlord().GetValue() == "" {
				isNeedPropertyUpdate = true
				people.Landlord = wrap.StringValue(&inviteeID)
			}

			if people.GetManager().GetValue() == "" {
				isNeedPropertyUpdate = true
				people.Manager = wrap.StringValue(&inviteeID)
			}

			if people.GetLandlordInvitation() == nil || people.GetLandlordInvitation().GetInviteeEmail() != inviteeEmail || !people.GetLandlordInvitation().GetPending() {
				isNeedPropertyUpdate = true
				sentAt, _ := ptypes.TimestampProto(time.Now())
				people.LandlordInvitation = &property.InvitationStatus{
					Pending:      true,
					Sent:         sentAt,
					InviteeEmail: inviteeEmail,
				}
			}

			if isNeedPropertyUpdate {
				p, err := PropertyService.UpdateAndVerify(ctx, &property.UpdateAndVerifyRequest{
					User:        req.GetInviter().GetId(),
					Property:    prop.GetProperty(),
					AdminVerify: &property.AdminVerify{},
				})
				if err != nil {
					log.Infof("CreateInvitation LANDLORD_INVITATION update property err %v\n", err)
					return nil, err
				}
				log.Infof("CreateInvitation LANDLORD_INVITATION update property success %v\n", p.GetProperty().GetId().GetValue())
			} else {
				log.Infof("CreateInvitation LANDLORD_INVITATION update not needed, property %+v\n", prop.GetProperty())
			}
		}
	case invitation.InvitationType_MANAGER_INVITATION: // if property in Listed state, set landlord as manager for a time
		prop, err := PropertyService.Get(ctx, &property.GetRequest{
			User: req.GetInviter().GetId(),
			Id:   req.GetProperty().GetId(),
		})
		if err != nil {
			return nil, err
		}

		isNeedPropertyUpdate := false
		people := prop.GetProperty().GetPeople()

		if !util.EqualStringValue(&req.GetInviter().Id, people.GetLandlord()) {
			return nil, ErrNotAllowed
		}

		if !util.EqualStringValue(&req.GetInviter().Id, people.GetManager()) {
			isNeedPropertyUpdate = true
			people.Manager = wrap.StringValue(&req.GetInviter().Id)
		}

		if people.GetManagerInvitation() == nil || people.GetManagerInvitation().GetInviteeEmail() != inviteeEmail || !people.GetManagerInvitation().GetPending() {
			isNeedPropertyUpdate = true
			sentAt, _ := ptypes.TimestampProto(time.Now())
			people.ManagerInvitation = &property.InvitationStatus{
				Pending:      true,
				Sent:         sentAt,
				InviteeEmail: inviteeEmail,
			}
		}

		if isNeedPropertyUpdate {
			p, err := PropertyService.UpdateAndVerify(ctx, &property.UpdateAndVerifyRequest{
				User:        req.GetInviter().GetId(),
				Property:    prop.GetProperty(),
				AdminVerify: &property.AdminVerify{},
			})
			if err != nil {
				log.Infof("CreateInvitation MANAGER_INVITATION update property err %v\n", err)
				return nil, err
			}
			log.Infof("CreateInvitation MANAGER_INVITATION update property success %+v\n", *p.GetProperty())
		} else {
			log.Infof("CreateInvitation MANAGER_INVITATION update not needed, property %+v\n", prop.GetProperty())
		}
	}

	// send event
	var embeds []*notification.EventEmbed
	embeds = append(embeds, &notification.EventEmbed{
		EmbedType:  messaging.EmbedType_EMBED_TYPE_INVITATION,
		EmbedValue: model.Id.Hex(),
	})
	embeds = append(embeds, &notification.EventEmbed{
		EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
		EmbedValue: model.PropertyID,
	})
	embeds = append(embeds, &notification.EventEmbed{
		EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
		EmbedValue: model.InviterID,
	})

	eventType := notification.EventType_EVENT_INVITATION_CREATE
	switch invitationType {
	case invitation.InvitationType_LANDLORD_INVITATION:
		eventType = notification.EventType_EVENT_INVITATION_CREATE_LANDLORD
	case invitation.InvitationType_MANAGER_TO_LANDLORD_INVITATION:
		eventType = notification.EventType_EVENT_INVITATION_CREATE_MANAGER_TO_LANDLORD
	case invitation.InvitationType_MANAGER_INVITATION:
		eventType = notification.EventType_EVENT_INVITATION_CREATE
	case invitation.InvitationType_TENANT_INVITATION:
		eventType = notification.EventType_EVENT_INVITATION_CREATE_TENANT
	}

	// to invitee
	_, err = server.notification.SendEvent(ctx, &notification.SendEventRequest{
		User:     model.InviteeID,
		Event:    eventType,
		Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_EMAILV2, notification.EventChannel_EVENT_CHANNEL_PLATFORM},
		Embeds:   embeds,
	})
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationActionResponse{
		Success: true,
		Info:    MakeInvitationInfo(model),
	}, nil
}

func declinedExistsInvitations(ctx context.Context, server *serverImpl, model *InvitationModel) error {
	models, err := server.repo.GetInvitations(ctx, GetInvitationsOptions{
		PropertyId:       model.PropertyID,
		InvitationType:   model.Type,
		InvitationStatus: invitation.InvitationStatus_PENDING,
	})
	if err != nil {
		log.WithError(err).Infof("isInvitationExist GetInvitations error, model: %+v", model)
		return err
	}

	log.Infof("declinedExistsInvitations GetInvitations len(models) = : %+v", len(models))
	if len(models) > 0 {
		for _, m := range models {
			err = server.repo.ChangeInvitationStatus(ctx, m.Id, invitation.InvitationStatus_DECLINED)
			if err != nil {
				log.WithError(err).Infof("declinedExistsInvitations change status to DECLINED error, model: %+v", m)
				return err
			}
			log.Infof("declinedExistsInvitations GetInvitations change status to DECLINED success %+v", m)
		}
	}

	return nil
}

func getInvitationInviteeInfo(ctx context.Context, invitee *invitation.UserResolved) (string, string, string, string) {
	inviteeFirstName := invitee.GetFirstName()
	inviteeLastName := invitee.GetLastName()
	inviteeID := invitee.GetId()
	inviteeEmail := invitee.GetEmail()

	// find invitee ID by email
	if inviteeID == "" && invitee.GetEmail() != "" {
		response, err := AccountService.GetByEmail(ctx, &account.GetByEmailRequest{Email: invitee.GetEmail()})
		if err != nil {
			log.WithError(err).Infof("getInvitationInviteeInfo AccountService.GetByEmail email: %v", invitee.GetEmail())
		} else {
			invitee.Id = response.GetAccount().GetIdValue()

			inviteeID = response.GetAccount().GetIdValue()
			inviteeFirstName = response.GetAccount().GetDetailsValue().GetFirstNameValue()
			inviteeLastName = response.GetAccount().GetDetailsValue().GetLastNameValue()
		}

		log.Infof(
			"getInvitationInviteeInfo find invitee ID by email InviteeID: %v ; inviteeEmail: %v\n",
			inviteeID,
			inviteeEmail,
		)
	}

	// find invitee email by account ID
	if inviteeEmail == "" && inviteeID != "" {
		resp, err := AccountService.Get(ctx, &account.GetRequest{Id: inviteeID})
		if err != nil {
			log.WithError(err).Infof("getInvitationInviteeInfo AccountService.Get ID: %v", inviteeID)
		}

		if resp.GetAccount().GetContactValue() != nil {
			inviteeEmail = resp.GetAccount().GetContactValue().GetEmailPrimaryValue()
			inviteeFirstName = resp.GetAccount().GetDetailsValue().GetFirstNameValue()
			inviteeLastName = resp.GetAccount().GetDetailsValue().GetLastNameValue()
		}

		log.Infof(
			"getInvitationInviteeInfo find invitee email by account ID Inviteeid: %v ; inviteeEmail: %v\n",
			inviteeID,
			inviteeEmail,
		)
	}

	return inviteeFirstName, inviteeLastName, inviteeID, inviteeEmail
}

func getInvitationMessageBody(req *invitation.InvitationResolution, id primitive.ObjectID) string {
	var bodyhi string
	if req.GetInvitee().GetId() == "" {
		bodyhi = fmt.Sprintf("Hello!")
	} else {
		bodyhi = fmt.Sprintf("Hello %s %s!", req.GetInvitee().GetFirstName(), req.GetInvitee().GetLastName())
	}

	propertyName := fmt.Sprintf("%s %s %s %s", req.GetProperty().GetBuilding(), req.GetProperty().GetFlat(), req.GetProperty().GetStreet(), req.GetProperty().GetCity())
	inviterName := fmt.Sprintf("%s %s", req.GetInviter().GetFirstName(), req.GetInviter().GetLastName())
	bodyurl := fmt.Sprintf("https://%s/property/view?id=%s", SiteUrl, req.GetProperty().GetId())
	bodydesc := fmt.Sprintf("%s has invited you to do the viewings for the following property %s\n%s", inviterName, propertyName, bodyurl)
	if req.GetMessage() != "" {
		bodydesc += fmt.Sprintf("\n\nHere is the message %s:\n%s\n", inviterName, req.GetMessage())
	}

	bodycontrol := fmt.Sprintf("To accept the invite please visit https://%s/profile/invites?id=%s", SiteUrl, id.Hex())

	body := fmt.Sprintf("%s\n%s\n\n%s\n", bodyhi, bodydesc, bodycontrol)
	return body
}

func (server *serverImpl) AcceptInvitation(ctx context.Context, in *invitation.InvitationActionRequest) (*invitation.InvitationActionResponse, error) {
	log.Infof("AcceptInvitation request: %+v\n", in)

	resp, err := server.changeInvitationStatus(ctx, in.GetId(), in.GetType(), in.GetUser(), in.GetEmail(), in.GetOtp(), invitation.InvitationStatus_ACCEPTED)
	if err != nil {
		log.Infof("AcceptInvitation changeInvitationStatus get error: %v\n", err)
		return resp, err
	}

	id, err := primitive.ObjectIDFromHex(in.Id)
	if err != nil {
		log.Infof("AcceptInvitation parsing id error: %v\n", err)
		return resp, err
	}

	inv, err := server.repo.GetInvitationById(ctx, id)
	if err != nil {
		log.Infof("AcceptInvitation GetInvitationById id: %v get error: %v\n", in.Id, err)
	}

	log.Infof("AcceptInvitation in.GetType() %v\n", in.GetType())
	switch inv.Type {
	case invitation.InvitationType_LANDLORD_INVITATION, invitation.InvitationType_MANAGER_TO_LANDLORD_INVITATION: // set inviter as Manager
		prop, err := PropertyService.Get(ctx, &property.GetRequest{
			User: in.User,
			Id:   inv.PropertyID,
		})
		if err == nil {
			log.Infof("AcceptInvitation LANDLORD_INVITATION found property: %v; user: %v\n", inv.PropertyID, in.User)

			people := prop.GetProperty().GetPeople()
			if !util.EqualStringValue(&inv.InviterID, people.GetManager()) {
				log.Infof("AcceptInvitation LANDLORD_INVITATION try to set Manager, update property: %v; manager: %v\n", inv.PropertyID, inv.InviterID)

				people.Manager = wrap.StringValue(&inv.InviterID)
				p, err := PropertyService.UpdateAndVerify(ctx, &property.UpdateAndVerifyRequest{
					User:        in.User,
					Property:    prop.GetProperty(),
					AdminVerify: &property.AdminVerify{},
				})
				if err != nil {
					log.Infof("AcceptInvitation LANDLORD_INVITATION set Manager, update property err %v\n", err)
				} else {
					log.Infof("AcceptInvitation LANDLORD_INVITATION set Manager, update property success %v\n", p.GetProperty().GetId().GetValue())
				}
			} else {
				log.Infof("AcceptInvitation LANDLORD_INVITATION no need set Manager with id %v\n", inv.InviterID)
			}
		} else {
			log.Infof("AcceptInvitation LANDLORD_INVITATION get property id: %v with user: %v get error: %v\n", inv.PropertyID, in.User, err)
		}
	}

	return resp, nil
}

func (server *serverImpl) DeclineInvitation(ctx context.Context, in *invitation.InvitationActionRequest) (*invitation.InvitationActionResponse, error) {
	return server.changeInvitationStatus(ctx, in.GetId(), in.GetType(), in.GetUser(), in.GetEmail(), in.GetOtp(), invitation.InvitationStatus_DECLINED)
}

func (server *serverImpl) DeclineInvitationById(ctx context.Context, in *invitation.InvitationActionRequest) (*invitation.InvitationActionResponse, error) {
	return server.declineInvitation(ctx, in.GetId())
}

func (server *serverImpl) InvitationStatus(ctx context.Context, in *invitation.InvitationStatusRequest) (*invitation.InvitationStatusResponse, error) {
	objId, err := primitive.ObjectIDFromHex(in.InvitationId)
	if err != nil {
		return nil, err
	}

	model, err := server.repo.GetInvitationById(ctx, objId)
	if err != nil {
		return nil, err
	}

	return &invitation.InvitationStatusResponse{
		Status: model.Status,
	}, nil
}

func MakeInvitationInfo(m *InvitationModel) *invitation.InvitationInfo {
	created, _ := ptypes.TimestampProto(m.CreatedAt)
	expires, _ := ptypes.TimestampProto(m.ExpiresAt)
	return &invitation.InvitationInfo{
		Id:               m.Id.Hex(),
		Type:             m.Type,
		InviterId:        m.InviterID,
		InviteeId:        m.InviteeID,
		InviteeEmail:     m.InviteeEmail,
		InviteeFirstName: m.InviteFirstName,
		InviteeLastName:  m.InviteLastName,
		PropertyId:       m.PropertyID,
		Body:             m.Body,
		CreatedAt:        created,
		ExpiresAt:        expires,
		Status:           m.Status,
		Offer:            m.Offer,
	}
}

func MakeInvitationListResponse(models []*InvitationModel) *invitation.InvitationListResponse {
	var infos []*invitation.InvitationInfo
	for _, m := range models {
		infos = append(infos, MakeInvitationInfo(m))
	}

	return &invitation.InvitationListResponse{
		Invitations: infos,
	}
}

func (server *serverImpl) InvitationListIncoming(ctx context.Context, in *invitation.InvitationListIncomingRequest) (*invitation.InvitationListResponse, error) {
	models, err := server.repo.ListIncomingInvitation(ctx, in.GetEmail(), in.GetProperty())
	if err != nil {
		return nil, err
	}
	return MakeInvitationListResponse(models), nil
}

func (server *serverImpl) InvitationListOutgoing(ctx context.Context, in *invitation.InvitationListOutgoingRequest) (*invitation.InvitationListResponse, error) {
	models, err := server.repo.ListOutgoingInvitation(ctx, in.GetUser())
	if err != nil {
		return nil, err
	}
	return MakeInvitationListResponse(models), nil
}

func (server *serverImpl) ListFilter(ctx context.Context, req *invitation.InvitationListFilterRequest) (*invitation.InvitationListFilterResponse, error) {
	models, count, err := server.FilterInvitations(ctx, req.GetParams())
	if err != nil {
		return nil, err
	}

	var infos []*invitation.InvitationInfo

	for _, m := range models {
		infos = append(infos, MakeInvitationInfo(m))
	}

	return &invitation.InvitationListFilterResponse{
		Invitations: infos,
		Total:       int32(count),
	}, nil
}

func (server *serverImpl) FilterInvitations(ctx context.Context, params *q.QueryParams) ([]*InvitationModel, int64, error) {
	query, opt, err := mongoquery.BuildFilter(InvitationModel{}, params, nil)
	if err != nil {
		return nil, 0, err
	}

	properties, err := server.repo.FindInvitations(ctx, query, opt)
	if err != nil {
		return nil, 0, err
	}

	total, err := server.repo.invitations.CountDocuments(ctx, query)
	return properties, total, err
}

func (server *serverImpl) GetInvitation(ctx context.Context, in *invitation.GetInvitationRequest) (*invitation.GetInvitationResponse, error) {
	objId, err := primitive.ObjectIDFromHex(in.Id)
	if err != nil {
		return nil, err
	}

	model, err := server.repo.GetInvitationById(ctx, objId)
	if err != nil {
		return nil, err
	}

	return &invitation.GetInvitationResponse{
		Info: &invitation.InvitationInfo{
			Id:               model.Id.String(),
			Type:             model.Type,
			InviterId:        model.InviterID,
			InviteeId:        model.InviteeID,
			InviteeEmail:     model.InviteeEmail,
			InviteeFirstName: model.InviteFirstName,
			InviteeLastName:  model.InviteLastName,
			PropertyId:       model.PropertyID,
			Body:             model.Body,
			Status:           model.Status,
			Offer:            model.Offer,
			Message:          model.Message,
		},
	}, nil
}

func (server *serverImpl) Update(ctx context.Context, in *invitation.UpdateRequest) (*invitation.UpdateResponse, error) {
	errResp := &invitation.UpdateResponse{
		Success: false,
	}

	if in.GetInvitation() == nil {
		errResp.Error = "empty invitation field not allowed"
		return errResp, grpcerr.New(grpcerr.InvalidData, errResp.Error)
	}

	if in.GetInvitation().GetId() == nil {
		errResp.Error = "empty invitation id not allowed"
		return errResp, grpcerr.New(grpcerr.InvalidData, errResp.Error)
	}

	id, err := primitive.ObjectIDFromHex(in.GetInvitation().GetIdValue())
	if err != nil {
		errResp.Error = "invalid invitation id"
		return errResp, grpcerr.New(grpcerr.InvalidData, errResp.Error)
	}

	model, err := server.repo.GetInvitationById(ctx, id)
	if err != nil {
		errResp.Error = err.Error()
		return errResp, err
	}

	if in.GetInvitation().GetInvitee() != nil {
		if in.GetInvitation().GetInviteeValue().GetId() != nil {
			model.InviteeID = in.GetInvitation().GetInviteeValue().GetIdValue()
		}
		if in.GetInvitation().GetInviteeValue().GetEmail() != nil {
			model.InviteeEmail = in.GetInvitation().GetInviteeValue().GetEmailValue()
		}
	}

	if in.GetInvitation().GetProcess() != nil {
		model.Process = in.GetInvitation().GetProcessValue()
	}

	err = server.repo.UpdateInvitation(ctx, model)
	if err != nil {
		errResp.Error = err.Error()
		return errResp, err
	}

	return &invitation.UpdateResponse{
		Success: true,
	}, nil
}

// CancelInvitationById to cancel (not) accepted invite by sender
func (server *serverImpl) CancelInvitationById(ctx context.Context, in *invitation.CancelInvitationByIdRequest) (*invitation.InvitationActionResponse, error) {
	var model *InvitationModel
	var err error

	oid, err := primitive.ObjectIDFromHex(in.Id)

	if len(in.Id) > 0 && err == nil {
		model, err = server.repo.GetInvitationById(ctx, oid)
	} else {
		return nil, ErrInvitationNotFound
	}
	if err != nil {
		return nil, err
	}

	if model.InviterID != in.User && !in.Admin {
		return nil, ErrInvitationNotFound
	}

	status := invitation.InvitationStatus_CANCELED
	if model.Type == invitation.InvitationType_MANAGER_INVITATION {
		switch model.Status {
		case invitation.InvitationStatus_PENDING:
			errResponse := &invitation.InvitationActionResponse{Error: invitation.InvitationActionError_UNKNOWN_ERROR, Success: false}
			err = server.repo.ChangeInvitationStatus(ctx, model.Id, status)
			if err != nil {
				return errResponse, err
			}

			model.Status = status

			if server.bpms != nil {
				c, err := server.bpms.NewPublishMessageCommand().
					MessageName("invitation-reacted").
					CorrelationKey(model.Id.Hex()).
					VariablesFromMap(
						map[string]interface{}{
							"accepted": false,
							"user":     model.InviteeID,
						})
				if err == nil {
					_, err = c.Send(context.Background())
					if err != nil {
						log.WithError(err).Error()
					}
				} else {
					log.WithError(err).Error()
				}
			}

			err := CancelInvitationInProperty(ctx, model)
			if err != nil {
				return nil, err
			}

			return &invitation.InvitationActionResponse{
				Success: true,
				Info:    MakeInvitationInfo(model),
			}, nil
		case invitation.InvitationStatus_ACCEPTED:
			err := CancelInvitationInProperty(ctx, model)
			if err != nil {
				return nil, err
			}

			return &invitation.InvitationActionResponse{
				Success: true,
				Info:    MakeInvitationInfo(model),
			}, nil

		default:
			return &invitation.InvitationActionResponse{
				Error:   invitation.InvitationActionError_UNKNOWN_ERROR,
				Success: false,
			}, nil
		}
	}
	return nil, ErrNotAllowed
}

// CancelInvitationsByProperty to cancel all invites of the property (by invitation type).
func (server *serverImpl) CancelInvitationsByProperty(ctx context.Context, in *invitation.CancelInvitationsByPropertyRequest) (*invitation.CancelInvitationsByPropertyResponse, error) {
	models, err := server.repo.GetInvitations(ctx, GetInvitationsOptions{
		PropertyId:     in.Property,
		InvitationType: in.InvitationType,
	})
	if err != nil {
		return nil, err
	}

	for _, i := range models {
		if i.InviterID == in.User {
			err = server.repo.ChangeInvitationStatus(ctx, i.Id, invitation.InvitationStatus_CANCELED)
			if err != nil {
				return nil, err
			}
			if i.Type == invitation.InvitationType_MANAGER_INVITATION &&
				i.Status == invitation.InvitationStatus_PENDING &&
				server.bpms != nil {
				c, err := server.bpms.NewPublishMessageCommand().
					MessageName("invitation-reacted").
					CorrelationKey(i.Id.Hex()).
					VariablesFromMap(
						map[string]interface{}{
							"accepted": false,
							"user":     i.InviteeID,
						})
				if err == nil {
					_, err = c.Send(context.Background())
					if err != nil {
						log.WithError(err).Error()
					}
				} else {
					log.WithError(err).Error()
				}
			}
			err = CancelInvitationInProperty(ctx, i)
			if err != nil {
				return nil, err
			}
		}
	}

	return &invitation.CancelInvitationsByPropertyResponse{Success: true}, nil
}

func CancelInvitationInProperty(ctx context.Context, in *InvitationModel) error {
	prop, err := PropertyService.Get(ctx, &property.GetRequest{
		Id:   in.PropertyID,
		User: in.InviterID,
	})
	if err != nil {
		return err
	}

	people := prop.GetProperty().GetPeople()

	people.Manager = people.GetLandlord()
	switch in.Type {
	case invitation.InvitationType_MANAGER_INVITATION:
		people.ManagerInvitation = &property.InvitationStatus{}
	case invitation.InvitationType_LANDLORD_INVITATION:
		// TODO add for landlord invitation
	}
	_, err = PropertyService.UpdateAndVerify(ctx, &property.UpdateAndVerifyRequest{
		User:        in.InviterID,
		Property:    prop.GetProperty(),
		AdminVerify: &property.AdminVerify{},
	})
	if err != nil {
		return err
	}

	return nil
}

func (server *serverImpl) ReferLandlord(ctx context.Context, in *invitation.ReferLandlordRequest) (*invitation.ReferLandlordResponse, error) {
	if in.InviteeEmail == "" || in.InviteeName == "" || in.InviterEmail == "" || in.InviterName == "" {
		return nil, ErrRequestEmpty
	}

	u, err := url.Parse(ReferLandlordGoogleSheetURL)
	if err != nil {
		return nil, err
	}
	qr := u.Query()
	qr.Set("invitee_email", in.InviteeEmail)
	qr.Set("invitee_name", in.InviteeName)
	qr.Set("inviter_email", in.InviterEmail)
	qr.Set("inviter_name", in.InviterName)

	u.RawQuery = qr.Encode()

	getResp, err := http.Get(u.String())
	if err != nil {
		return nil, err
	}
	if getResp.StatusCode != http.StatusOK {
		return nil, ErrBadHTTPResponseStatus
	}
	return &invitation.ReferLandlordResponse{
		Success: true,
	}, nil
}
