package impl

import (
	"context"
	"time"

	"gitlab.okta-solutions.com/mashroom/backend/common/api/grpcerr"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo/options"

	invitation "gitlab.okta-solutions.com/mashroom/backend/invitation/api"
)

type InvitationModel struct {
	Id              primitive.ObjectID          `bson:"_id,omitempty"`
	Type            invitation.InvitationType   `bson:"type"`
	InviterID       string                      `bson:"inviter_id"`
	InviteeID       string                      `bson:"invitee_id"`
	InviteeEmail    string                      `bson:"invitee_email"`
	InviteFirstName string                      `bson:"invite_first_name"`
	InviteLastName  string                      `bson:"invite_last_name"`
	Otp             string                      `bson:"otp"`
	PropertyID      string                      `bson:"property_id"`
	Status          invitation.InvitationStatus `bson:"status"`
	Body            string                      `bson:"body"`
	CreatedAt       time.Time                   `bson:"created_at"`
	ExpiresAt       time.Time                   `bson:"expires_at"`
	Process         int64                       `bson:"process"`
	Offer           string                      `bson:"offer"`
	CallbackUrl     string                      `bson:"callback_url"`
	Message         string                      `bson:"message"`
}

type GetInvitationsOptions struct {
	PropertyId       string
	InvitationType   invitation.InvitationType
	InvitationStatus invitation.InvitationStatus
}

func (repo *Repo) CreateInvitation(ctx context.Context, model *InvitationModel) (primitive.ObjectID, error) {
	model.CreatedAt = time.Now()

	_, err := repo.invitations.InsertOne(ctx, model)

	return model.Id, err
}

func (repo *Repo) GetInvitation(ctx context.Context, invitationType invitation.InvitationType, inviteeEmail string, otp string) (*InvitationModel, error) {
	result := InvitationModel{}
	query := bson.M{"type": invitationType, "invitee_email": inviteeEmail, "otp": otp}

	err := repo.invitations.FindOne(ctx, query).Decode(&result)
	if err != nil {
		return nil, err
	}

	return &result, err
}

func (repo *Repo) ListOutgoingInvitation(ctx context.Context, user string) ([]*InvitationModel, error) {
	var result []*InvitationModel

	query := bson.M{
		"inviter_id": user,
	}

	cur, err := repo.invitations.Find(ctx, query)
	if err != nil {
		return nil, err
	}

	err = cur.All(ctx, &result)

	return result, err
}

func (repo *Repo) ListIncomingInvitation(ctx context.Context, email, property string) ([]*InvitationModel, error) {
	var result []*InvitationModel

	query := bson.M{
		"invitee_email": email,
	}

	if property != "" {
		query["property_id"] = property
	}

	cur, err := repo.invitations.Find(ctx, query)
	if err != nil {
		return nil, err
	}

	err = cur.All(ctx, &result)

	return result, err
}

func (repo *Repo) GetInvitationById(ctx context.Context, invitationId primitive.ObjectID) (*InvitationModel, error) {
	result := InvitationModel{}

	err := repo.invitations.FindOne(ctx, bson.M{"_id": invitationId}).Decode(&result)

	return &result, err
}

func (repo *Repo) ChangeInvitationStatus(ctx context.Context, id primitive.ObjectID, status invitation.InvitationStatus) error {
	count, err := repo.invitations.CountDocuments(ctx, bson.M{"_id": id})
	if err != nil {
		return err
	}

	if count == 0 {
		return grpcerr.New(grpcerr.NotFound, "invitation does not exist")
	}

	_, err = repo.invitations.UpdateOne(ctx, bson.M{"_id": id}, bson.M{
		"$set": bson.M{
			"status": status,
		},
	})

	return err
}

func (repo *Repo) GetInvitations(ctx context.Context, opts GetInvitationsOptions) ([]*InvitationModel, error) {
	var result []*InvitationModel

	query := bson.M{
		"type":        opts.InvitationType,
		"property_id": opts.PropertyId,
		"status": bson.M{"$in": []int32{
			int32(invitation.InvitationStatus_PENDING),
			int32(invitation.InvitationStatus_ACCEPTED),
		}},
	}

	if opts.InvitationStatus != invitation.InvitationStatus_UNKNOWN_STATUS {
		query["status"] = opts.InvitationStatus
	}

	cur, err := repo.invitations.Find(ctx, query)
	if err != nil {
		return nil, err
	}

	err = cur.All(ctx, &result)

	return result, err
}

func (repo *Repo) UpdateInvitation(ctx context.Context, model *InvitationModel) error {
	_, err := repo.invitations.UpdateOne(ctx, bson.M{"_id": model.Id}, model)

	return err
}

func (repo *Repo) FindInvitations(ctx context.Context, m bson.M, opts ...*options.FindOptions) ([]*InvitationModel, error) {
	var results []*InvitationModel

	cur, err := repo.invitations.Find(ctx, m, opts...)
	if err != nil {
		log.WithField("query", m).Info("trace query")
		return nil, err
	}

	err = cur.All(ctx, &results)

	return results, err
}
