package impl

import (
	"context"
	"fmt"
	"time"

	"github.com/globalsign/mgo/bson"
	"gitlab.okta-solutions.com/mashroom/backend/common/api/grpcerr"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"

	"gitlab.okta-solutions.com/mashroom/backend/common/generic/env"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
)

const (
	InvitationsCollection        = "invitations"
	ContactValidationsCollection = "contact-validations"
	PasswordResetsCollection     = "password-resets"
)

var (
	MongoUrl      = env.ResolveString("MONGO_URL", "mongodb:27017")
	MongoUsername = env.ResolveString("MONGO_USERNAME", "invitation-data")
	MongoPassword = env.ResolveString("MONGO_PASSWORD", "invitation-data")
	MongoAuthDb   = env.ResolveString("MONGO_AUTH", "admin")
	MongoDb       = env.ResolveString("MONGO_DATABASE", "invitation-data")
)

type Repo struct {
	db                 *mongo.Database
	invitations        *mongo.Collection
	contactValidations *mongo.Collection
	passwordResets     *mongo.Collection
}

func (repo *Repo) ChangeDatabase(db string) {
	repo.db = repo.db.Client().Database(db)
	repo.invitations = repo.db.Collection(InvitationsCollection)
	repo.contactValidations = repo.db.Collection(ContactValidationsCollection)
	repo.passwordResets = repo.db.Collection(PasswordResetsCollection)
}

func (repo *Repo) Close() {
	_ = repo.db.Client().Disconnect(context.Background())
}

func NewRepo() (*Repo, error) {
	return NewRepoWithUrl(MongoUrl)
}

func NewRepoWithUrl(mongoUrl string) (*Repo, error) {
	log.Info("Connecting to mongo")

	client, err := mongo.NewClient(options.Client().ApplyURI(fmt.Sprintf("mongodb://%s", mongoUrl)).SetAuth(options.Credential{
		AuthSource:  "admin",
		Username:    MongoUsername,
		Password:    MongoPassword,
		PasswordSet: true,
	}))
	if err != nil {
		return nil, err
	}

	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
	defer cancel()

	if err := client.Connect(ctx); err != nil {
		return nil, err
	}

	db := client.Database(MongoDb)
	log.Info("Mongo connected.")
	return &Repo{
		db:                 db,
		invitations:        db.Collection(InvitationsCollection),
		contactValidations: db.Collection(ContactValidationsCollection),
		passwordResets:     db.Collection(PasswordResetsCollection),
	}, nil
}

func StringToObjectId(s string) (primitive.ObjectID, error) {
	if bson.IsObjectIdHex(s) {
		return primitive.ObjectIDFromHex(s)
	}

	return primitive.NewObjectID(), grpcerr.New(grpcerr.InvalidData, "string format incorrect")
}
